package org.vaadin.desktopnotifications;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import org.vaadin.desktopnotifications.shared.DesktopNotifierClientRpc;
import org.vaadin.desktopnotifications.shared.DesktopNotifierServerRpc;
import org.vaadin.desktopnotifications.shared.DesktopNotifierState;
import org.vaadin.desktopnotifications.shared.SharedNotification;

import com.vaadin.server.AbstractExtension;
import com.vaadin.server.Resource;

/**
 * Server side component for the VDesktopNotifier widget.
 */
public class DesktopNotifier extends AbstractExtension implements
		DesktopNotifierServerRpc {
	private static final long serialVersionUID = -4935586799603427503L;

	public interface SettingsListener {

		/**
		 * @param isSupportedByBrowser
		 * @param isAllowed
		 *            <code>null</code> if the user hasn't given an answer yet
		 * @param isDisallowed
		 *            <code>null</code> if the user hasn't given an answer yet
		 */
		void notificationSettingsChanged(Boolean isSupportedByBrowser,
				Boolean isAllowed, Boolean isDisallowed);

	}

	private Boolean isSupportedByBrowser = null;
	private Boolean isAllowed;
	private Boolean isDisallowed;

	private final List<SettingsListener> listeners = new ArrayList<DesktopNotifier.SettingsListener>();

	public DesktopNotifier() {
		registerRpc(this);
	}

	/**
	 * Checks whether the browser supports desktop notifications as a feature.
	 * 
	 * @throws MethodWasCalledBeforeClientRoundtripException
	 *             if the method has been called before there has been a
	 *             server&rarr;client&rarr;server roundtrip.
	 */
	public boolean notificationsAreSupportedByBrowser()
			throws MethodWasCalledBeforeClientRoundtripException {
		if (isSupportedByBrowser == null) {
			throw new MethodWasCalledBeforeClientRoundtripException();
		} else {
			return isSupportedByBrowser;
		}
	}

	/**
	 * Checks whether the user has allowed desktop notifications.
	 * 
	 * @return <code>true</code> iff the user has given an explicit permission
	 *         for the server to send desktop notifications. <code>false</code>
	 *         is returned if the user hasn't answered to the permission prompt,
	 *         or if the user has explicitly disallowed notifications.
	 * @throws MethodWasCalledBeforeClientRoundtripException
	 *             if the method was called before a server-client roundtrip was
	 *             completed.
	 */
	public boolean notificationsAreAllowedByUser()
			throws MethodWasCalledBeforeClientRoundtripException {
		if (browserHasSentSupportInformation()) {
			return (isAllowed != null) ? isAllowed : false;
		} else {
			throw new MethodWasCalledBeforeClientRoundtripException();
		}
	}

	private boolean browserHasSentSupportInformation() {
		return isSupportedByBrowser != null;
	}

	/**
	 * Checks whether the user has disallowed desktop notifications
	 * 
	 * @return <code>true</code> iff the user has explicitly denied the server
	 *         to send desktop notifications. <code>false</code> is returned if
	 *         the user hasn't answered to the permission prompt, or if the user
	 *         has explicitly allowed notifications.
	 * @throws MethodWasCalledBeforeClientRoundtripException
	 *             if the method was called before a server-client roundtrip was
	 *             completed.
	 */
	public boolean notificationsAreDisallowedByUser()
			throws MethodWasCalledBeforeClientRoundtripException {
		if (browserHasSentSupportInformation()) {
			return (isDisallowed != null) ? isDisallowed : false;
		} else {
			throw new MethodWasCalledBeforeClientRoundtripException();
		}
	}

	/**
	 * Show a plaintext notification
	 * <p/>
	 * <strong>Note:</strong> This method does nothing if
	 * 
	 * <ul>
	 * <li>{@link #notificationsAreSupportedByBrowser()} is <code>false</code>
	 * OR
	 * <li>{@link #notificationsAreDisallowedByUser()} is <code>true</code> OR
	 * <li>{@link #notificationsAreAllowedByUser()} is <code>false</code>
	 * </ul>
	 * 
	 * @param iconUrl
	 * @param header
	 * @param body
	 */
	public SharedNotification showNotification(final String iconUrl, final String header,
			final String body) {
	    final SharedNotification notification = new SharedNotification(iconUrl, header, body);
	    getRpcProxy(DesktopNotifierClientRpc.class).showNotification(notification);
	    return notification;
	}

	/**
	 * Show a plaintext notification
	 * <p/>
	 * <strong>Note:</strong> This method does nothing if
	 * 
	 * <ul>
	 * <li>{@link #notificationsAreSupportedByBrowser()} is <code>false</code>
	 * OR
	 * <li>{@link #notificationsAreDisallowedByUser()} is <code>true</code> OR
	 * <li>{@link #notificationsAreAllowedByUser()} is <code>false</code>
	 * </ul>
	 * 
	 * @param iconResource
	 * @param header
	 * @param body
	 */
	public SharedNotification showNotification(final Resource iconResource,
			final String header, final String body) {
		setResource(DesktopNotifierState.ICON, iconResource);
		final SharedNotification notification = new SharedNotification(null, header, body);
		notification.setId(new Random().nextInt());
		getRpcProxy(DesktopNotifierClientRpc.class).showNotification(notification);
		return notification;
	}
	
	/**
	 * Hide an existing plaintext notification.
	 * 
	 * @since
	 * @param notification
	 */
	public void hideNotification(final SharedNotification notification) {
	    if (notification != null) {
        	getRpcProxy(DesktopNotifierClientRpc.class).hideNotification(notification);
	    }
	}

	/**
	 * Prompts the user with a request for permission for showing desktop
	 * notifications.
	 * 
	 * @see #setFeatureDescriptionText(String)
	 */
	public void requestPermission() {
		getRpcProxy(DesktopNotifierClientRpc.class).requestPermission();
	}

	@Override
	public DesktopNotifierState getState() {
		return (DesktopNotifierState) super.getState();
	}

	/**
	 * Change the text that is shown to the user in the popup that will be
	 * presented before the user can decide whether she allows desktop
	 * notifications.
	 * 
	 * @param text
	 */
	public void setFeatureDescriptionText(final String text) {
		getState().text = text;
	}

	public void addListener(final SettingsListener listener) {
		if (listener != null) {
			listeners.add(listener);
		}
	}

	public void removeListener(final SettingsListener listener) {
		listeners.remove(listener);
	}

	@Override
	public void notificationsAreAllowedAndSupported(final boolean allowed,
			final boolean supported) {
		isAllowed = allowed;
		isDisallowed = !allowed;
		isSupportedByBrowser = supported;

		fireListeners();
	}

	private void fireListeners() {
		for (final SettingsListener listener : listeners) {
			listener.notificationSettingsChanged(isSupportedByBrowser,
					isAllowed, isDisallowed);
		}
	}

	@Override
	public void notificationsAreSupported(final boolean supported) {
		isSupportedByBrowser = supported;
		fireListeners();
	}
}
